/++
Модуль, представляющий класс $(LINK2 #Камера, Камера) - 
3D-объект, из которого осуществляется просмотр 3D-сцены. 

Одновременно в приложении может быть определено несколько камер,
но действует только та, на которую ссылается поле $(B камера) $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджера)
+/
module ddd.camera;

import std.string;
//import derelict.opengl.gl;

import ddd.dvigimyi_object, ddd.matrix, ddd.quaternion, ddd.vector;
import ddd.zavisimost.graphika_gl;
import my_utils.journal;

private enum имя_модуля = "камера";

/// класс, определяющий камеру - 3D-объект, из которого осуществляется просмотр 3D-сцены.
class Камера : ДвижимыйОбъект {
    
    private Матрица4 матрица;

    /// При выводе на экран ничего не происходит
    override void рисовать(){};

    /// Конструктор, принимающий на вход имя
    public this(string имя) {
        super(имя);
        //матрица = Матрица4.НУЛЕВАЯ_МАТРИЦА;
        матрица = Матрица4(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }   

    /++
    Установка направления камеры в соответствии с кватернионом поворота родительского 
    $(LINK2 mesto.html#Место, Места). Данные о положении камеры игнорируются.
    +/
    void установить_направление_камеры(bool обновить) {
        if (обновить) 
            обновить_матрицу();
        Матрица4 матрица_поворота = Матрица4(матрица.получить_матрицу_3x3());
        gl_загрузить_матрицу(&(матрица_поворота.m[0]));
    }
    
    /++
    Установка направления и положения камеры в соответствии с родительским
    $(LINK2 mesto.html#Место, Местом).
    +/
    void установить_камеру(bool обновить) {
        // Текущее положение в матричном виде
        if (обновить) 
            обновить_матрицу();
        gl_загрузить_матрицу(&(матрица.m[0]));
        //glLoadMatrixf(&(mat.m[0]));
        
    }
    
    private void обновить_матрицу() {
        auto обратная_матрица = родитель.получить_унаследованную_матрицу();
        матрица = обратная_матрица.обратная();
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("Матрица камеры %s", матрица.toString()));
        }
}
