/++
Модуль, определяющий класс $(LINK2 #ЭлементИнфопанели, ЭлементИнфопанели), который позволяет 
добавлять текстуру и другие элементы на существующую $(LINK2 infopanel.html#Инфопанель, Инфопанель).
+/
module ddd.element_infopaneli;

import std.conv, std.string, std.json;
import ddd.mesh, ddd.texture, ddd.ddd_manager, ddd.types;
import ddd.zavisimost.graphika_gl;

import std.stdio; //, std.string;

import my_utils.journal;

enum имя_модуля = "э-т_инфопанели";

/++
Элемент инфопанели в json-файле:
$(UL  
$(LI  $(B имя) (строковый) - имя инфопанели в программе) 
$(LI  $(B контейнер) (bool) - является ли этот элемент контейнером для других элементов)
$(LI  $(B координаты) (массив из 4-х значений типа float) - координаты элемента 
относительно расположения и размеров всей инфопанели или родительского контейнера: [
координата x левой границы, координата y нижней границы, ширина, высота])  
$(LI  $(B текстура) (строковый) - название ресурса текстуры, т.е. имя файла текстуры без расширения)
$(LI  $(B координаты_текстуры) (массив из 4-х значений типа float) - текстурные координаты)
$(LI  $(B элементы) - массив объектов $(LINK2 element_infopaneli.html#ЭлементИнфопанели, ЭлементИнфопанели))  
)
+/
enum JSON_TYPE[string] ЭлементИнфопанели_JSON = [
    "имя": JSON_TYPE.STRING,
    "контейнер": JSON_TYPE.FALSE,
    "координаты": JSON_TYPE.ARRAY,
    "текстура": JSON_TYPE.STRING,
    "координаты_текстуры": JSON_TYPE.ARRAY,
    "элементы": JSON_TYPE.ARRAY
    ];

private enum float[] ВЕРШИНЫ_МЕША = [0,0,0,1,1,0,0,1,0,0,0,0,1,0,0,1,1,0];
private enum ushort[] ИНДЕКСЫ_МЕША = [0,1,2,3,4,5];
private enum size_t[] ИНДЕКСЫ_ТЕКСТУРНЫХ_КООРДИНАТ = [0,3,2,1,0,1,0,3,2,3,2,1];

/++
Класс, позволяющий
добавлять текстуру и другие элементы на существующую $(LINK2 infopanel.html#Инфопанель, Инфопанель).

В отличие от $(LINK2 infopanel.html#Инфопанель, Инфопанели), координаты объектов этого класса,
передаваемые в конструкторе, всегда относительные.
+/
class ЭлементИнфопанели {
    protected {
        bool включено = true;
        float относительный_x1, относительный_y1, относительный_x2, относительный_y2;
        float относительная_ширина, относительная_высота;
        float[4] координаты_текстуры; // Инициализируется в 0
        int x,y, ширина, высота;
        string имя;
        bool _является_контейнером;
        ЭлементИнфопанели[string] элементы;
        Текстура текстура = null;
        Меш меш = null;
        ЭлементИнфопанели _родитель = null;
    }
    
    private static float[12] координаты_текстуры_меша; // Временное место передачи координат

    /// Возвращает элемент инфопанели, являющийся родительским к этому элементу.
    public @property ЭлементИнфопанели родитель() { return _родитель;}

    /// Присваивает родительский элемент этому элементу.
    public @property void родитель(ЭлементИнфопанели родитель0){
        _родитель = родитель0;
        if (родитель0 !is null)
            рассчитать_абсолютные_координаты(родитель0.абсолютные_размеры); 
    }

    /++
    Конструктор, принимающий минимальное количество параметров.

    Элемент, созданный этим конструктором, будет занимать всю площадь инфопанели или родительского контейнера.

    Params:
        имя0 = Имя создаваемого элемента.
        будет_контейнером = Возможно ли вставлять в этот элемент другие элементы.
    +/
    this(string имя0, bool будет_контейнером) {
        this(имя0, будет_контейнером, 0, 0, 1, 1);
    }
    
    /++
    Конструктор, учитывающий координаты в относительных единицах.

    Отрицательные координаты x0 и/или y0 задают положение от конца инфопанели или 
    родительского контейнера (x0 от правого края, y0 от верхнего).
    

    Params:
        имя0 = Имя создаваемого элемента.
        будет_контейнером = Возможно ли вставлять в этот элемент другие элементы.
        x0 = Координата x левой границы элемента.
        y0 = Координата y нижней границы элемента.
        ширина0 = Ширина элемента.
        высота0 = Высота элемента.
    +/
    this(string имя0, bool будет_контейнером, float x0, float y0, float ширина0, float высота0) {
        имя = имя0;
        _является_контейнером = будет_контейнером;
        инициализация_координат_текстуры();
        //writeln("Создание элемента инфопанели:", имя0, x0, y0, ширина0, высота0);
        задать_координаты(x0, y0, ширина0, высота0);
    }

    /++
    Функция, задающая координаты элемента в относительных единицах.


    Отрицательные координаты x0 и/или y0 задают положение от конца инфопанели или 
    родительского контейнера (x0 от правого края, y0 от верхнего).
    

    Params:
        x0 = Координата x левой границы элемента.
        y0 = Координата y нижней границы элемента.
        ширина0 = Ширина элемента.
        высота0 = Высота элемента.
    +/
    public void задать_координаты(float x0, float y0, float ширина0, float высота0) {
        /*
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("координаты для элемента инфопанели %s x0=%s, y0=%s, ширина0=%s, высота0=%s",
                   имя, x0, y0, ширина0, высота0));
        */
        относительный_x1 = x0>=0? x0 : 1+x0;
        относительный_y1 = y0>=0? y0 : 1+y0;
        относительная_ширина = ширина0;
        относительная_высота = высота0;
        относительный_x2 = относительный_x1 + относительная_ширина;
        относительный_y2 = относительный_y1 + относительная_высота;

        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("установка координат элемента инфопанели %s отн_x1=%s, отн_y1=%s, отн_x2=%s, отн_y2=%s",
                   имя, относительный_x1, относительный_y1, относительный_x2, относительный_y2)); 
    }

    /++
    Функция, вычисляющая абсолютные координаты элемента в пикселях.
    

    Params:
        ширина_родителя = Ширина родительского элемента-контейнера в пикселях.
        высота_родителя = Высота родительского элемента-контейнера в пикселях.
    +/
    public void рассчитать_абсолютные_координаты(int ширина_родителя, int высота_родителя) {
        this.x = to!int(относительный_x1 * ширина_родителя);
        this.y = to!int(относительный_y1 * высота_родителя);
        this.ширина = to!int(относительная_ширина * ширина_родителя);
        this.высота = to!int(относительная_высота * высота_родителя);
    }

    
    /++
    Функция, вычисляющая абсолютные координаты элемента в пикселях.
    

    Params:
        ширина_и_высота = Ширина и высота родительского элемента-контейнера в формате 
        $(LINK2 types.html#Координаты2D, Координаты2D).
    +/
    public void рассчитать_абсолютные_координаты(Координаты2D!(int) ширина_и_высота) {
        рассчитать_абсолютные_координаты(ширина_и_высота.x, ширина_и_высота.y);
    }

    /++
    Функция, возвращающая абсолютные размеры элемента в пикселях в фомате 
    $(LINK2 types.html#Координаты2D, Координаты2D).
    +/
    public @property Координаты2D!int абсолютные_размеры() {
        return Координаты2D!int(ширина, высота);
    }

    /++
    Функция, позволяющая задать абсолютные размеры элемента в пикселях.

    Полученные размеры в случае необходимости будут обрезаны, чтобы не выйти за пределы
    инфопанели или родительского элемента.
    
    Params:
        ширина0 = Ширина элемента в пикселях.
        высота0 = Высота элемента в пикселях.
    +/
    public void задать_абсолютные_размеры(int ширина0, int высота0) {
        auto размеры_родителя = _родитель.абсолютные_размеры;
        int ширина_родителя = размеры_родителя.x, высота_родителя = размеры_родителя.y;
        if ((ширина_родителя <= 0) || (высота_родителя <= 0) )
            return;
            
        if (ширина_родителя >= (ширина0 + относительный_x1 * ширина_родителя)) {
            относительная_ширина = to!float(ширина0)/ширина_родителя;
            относительный_x2 = относительный_x1 + относительная_ширина;
            this.ширина = ширина0;
        }
        else {
            относительный_x2 = 1;
            относительная_ширина = относительный_x2 - относительный_x1;
            this.ширина = to!int(относительная_ширина * ширина_родителя);
            координаты_текстуры[2] = относительная_ширина / (ширина0 / ширина_родителя);
        }
        
        if (высота_родителя >= (высота0 + относительный_y1 * высота_родителя)) {
            относительная_высота = to!float(высота0) / высота_родителя;
            относительный_y2 = относительный_y1 + относительная_высота;
            this.высота = высота0;
        }
        else {
            относительный_y2 = 1;
            относительная_высота = относительный_y2 - относительный_y1;
            this.высота = to!int(относительная_высота * высота_родителя);
            координаты_текстуры[3] = относительная_высота / (высота0 / высота_родителя);
        }
    }    

    /++
    Функция, задающая текстурные координаты для текстуры этого элемента.
    
    Params:
        x1 = Текстурная координата x левого нижнего угла части текстуры.
        y1 = Текстурная координата y левого нижнего угла части текстуры.
        x2 = Текстурная координата x правого верхнего угла части текстуры.
        y2 = Текстурная координата y правого верхнего угла части текстуры.
    +/
    public void задать_координаты_текстуры(float x1, float y1, float x2, float y2) {
        координаты_текстуры[0]=x1;
        координаты_текстуры[1]=y1;
        координаты_текстуры[2]=x2;
        координаты_текстуры[3]=y2;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("Присваивание текстурных координат %s координаты_текстуры=%s",
                   имя, координаты_текстуры)); 
        if (меш) 
            меш.изменить_координаты_текстур(_преобразовать_координаты_текстур());
    }


    /++
    Функция, задающая текстурные координаты для текстуры этого элемента.
    
    Params:
        координаты_текстуры0 = Текстурные координаты в виде массива из 4-х элементов
        [x1, y1, x2, y2].
    +/
    public void задать_координаты_текстуры(float[4] координаты_текстуры0) {
        
        координаты_текстуры = координаты_текстуры0;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("Присваивание текстурных координат %s координаты_текстуры=%s",
                   имя, координаты_текстуры)); 
        if (меш) 
            меш.изменить_координаты_текстур(_преобразовать_координаты_текстур());    
    }

    protected void инициализация_координат_текстуры() {
        координаты_текстуры[0]=0;
        координаты_текстуры[1]=0;
        координаты_текстуры[2]=1;
        координаты_текстуры[3]=1;
    }

    private float[] _преобразовать_координаты_текстур() {
        for (size_t индекс=0; индекс<ИНДЕКСЫ_ТЕКСТУРНЫХ_КООРДИНАТ.length; индекс++)
            координаты_текстуры_меша[индекс] = координаты_текстуры[ИНДЕКСЫ_ТЕКСТУРНЫХ_КООРДИНАТ[индекс]];
        return координаты_текстуры_меша;
    }
    
    protected void создать_меш() {
        float[] координаты_вершин = ВЕРШИНЫ_МЕША; 
        ushort[] индексы = ИНДЕКСЫ_МЕША;
        //float[] координаты_текстур = [координаты_текстуры[0], координаты_текстуры[3],
        //                      координаты_текстуры[2], координаты_текстуры[1],
        //                      координаты_текстуры[0], координаты_текстуры[1],
        //                      координаты_текстуры[0], координаты_текстуры[3],
        //                      координаты_текстуры[2], координаты_текстуры[3],
        //                      координаты_текстуры[2], координаты_текстуры[1]];
        //меш = new Меш(координаты_вершин, координаты_текстур, null, индексы);                      
        меш = new Меш(координаты_вершин, _преобразовать_координаты_текстур(), null, индексы);
    }

    /// Функция, позволяющая включить или выключить видимость этого элемента инфопанели.
    public void задать_видимость(bool включено0) {
        включено = включено0;
    }

    /// Функция, возвращающая значение видимости этого элемента инфопанели.
    public @property bool получить_видимость() {
        return включено;
    }

    /// Функция, позволяющая переключить видимость этого элемента инфопанели.
    public void поменять_видимость() {
        включено = !включено;
    }

    /// Функция, возвращающая значение имени этого элемента инфопанели.
    public @property string получить_имя() {
        return имя;
    }

    /// Функция, назначающая текстуру этому элементу инфопанели.
    public void задать_текстуру(Текстура текстура0) {
        текстура = текстура0;
        //if (меш is null) создать_меш();
    }

    /++
    Вывод этого элемента инфопанели на экран. Также выводятся все дочерние элементы, 
    если они присутствуют, и если этот элемент является контейнером. Если видимость отключена, 
    то функция не выполняет никаких действий.
    +/
    public void рисовать() {
        if (!включено) return;
        if ((меш is null) && (текстура !is null)) создать_меш();
        gl_подготовить_рисование_элемента_инфопанели(относительный_x1, относительный_y1, 
                                                     относительная_ширина, относительная_высота);
        if (текстура !is null) {   //&& (меш !is null)) 
            текстура.рисовать();
            меш.рисовать();
        }
        if (_является_контейнером) 
            foreach (элемент; элементы.byValue()) 
                элемент.рисовать();
        gl_восстановить_матрицу();
    }

    /// Функция, возвращающая возможность для этого элемента содержать дочерние элементы.
    public @property bool является_контейнером() {
        return _является_контейнером;
    }

    /++
    Присоединение переданного элемента в контейнер этого элемента.

    Если этот элемент не является контейнером, или у переданного элемента уже есть родитель,
    то функция не выполняет никаких действий.

    Params:
        элемент = Элемент, который требуется добавить в контейнер этого элемента.
    +/
    public void присоединить_элемент(ЭлементИнфопанели элемент) {
        if (!является_контейнером) return;
        if (элемент.родитель !is null) return;
        элементы[элемент.получить_имя()] = элемент;
        элемент.родитель = this;
    }

    /++
    Отсоединение дочернего элемента от этого элемента инфопанели.


    Params:
        элемент = отсоединяемый элемент


    Returns:
        Отсоединённый элемент, если он присуствовал в этом элементе, null, если нет.
    +/
    public ЭлементИнфопанели отсоединить_элемент(ЭлементИнфопанели элемент) {
        return отсоединить_элемент(элемент.получить_имя());
    }

    /++
    Отсоединение дочернего элемента по его имени от этого элемента инфопанели.


    Params:
        имя_элемента = имя отсоединяемого элемента


    Returns:
        Отсоединённый элемент, если он присуствовал в этом элементе, null, если нет.
    +/
    public ЭлементИнфопанели отсоединить_элемент(string имя_элемента) {
        if (является_контейнером && (имя_элемента in элементы)) {
            ЭлементИнфопанели элемент = элементы[имя_элемента];
            элементы.remove(имя_элемента);
            элемент.родитель = null;
            return элемент;
        }
        else
            return null;
    }

    /++
    Получение ссылки на дочерний элемент по его имени.


    Params:
        имя_элемента = имя искомого элемента


    Returns:
        Искомый элемент, если он присуствовал в этом элементе, null, если нет.
    +/
    public ЭлементИнфопанели получить_элемент(string имя_элемента) {
        if (является_контейнером && (имя_элемента in элементы)) 
            return элементы[имя_элемента];
        else
            return null;
    }

    /++
    Функция генерации нового элемента инфопанели (возможно, с присоединёнными элементами) из JSON-объекта.


    Params:
        json_значение0 = JSON-объект с информацией об элементе 
        (см. $(LINK2 #ЭлементИнфопанели_JSON, ЭлементИнфопанели_JSON)).
        менеджер0 = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)


    Returns:
        Созданный элемент инфопанели, если генерация прошла успешно, null в противном случае.
    +/
    static public ЭлементИнфопанели из_JSON(JSONValue json_значение0, DDDМенеджер менеджер0) {
        JSONValue[string] json_значение1;
        string имя;
        bool будет_контейнером = false;
        JSONValue[] координаты;
        JSONValue[] координаты_текстуры;
        string имя_текстуры;
        ЭлементИнфопанели результат;

        try {
            json_значение1 = json_значение0.object;
            имя = json_значение1["имя"].str;
            координаты = json_значение1["координаты"].array;
        }
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, "Ошибка получения данных для создания ЭлементИнфопанели");
            return null;
        }
        if ("контейнер" in json_значение1) {
            if (json_значение1["контейнер"].type == JSON_TYPE.TRUE)
                будет_контейнером = true;
            else 
                if (json_значение1["контейнер"].type != JSON_TYPE.FALSE)
                    в_журнал(имя_модуля, УровниЖурнала.Ошибка, 
                       "Ошибка определения наличия контейнера для ЭлементИнфопанели с именем " ~ имя);
        }
        try {
            результат = new ЭлементИнфопанели(имя, будет_контейнером, координаты[0].floating, 
                            координаты[1].floating, координаты[2].floating, координаты[3].floating);
        }
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, "Ошибка создания ЭлементИнфопанели с именем " ~ имя);
            return null;
        }
        if ("текстура" in json_значение1) {
            Текстура текстура;
            try {
                имя_текстуры = json_значение1["текстура"].str;
                текстура = менеджер0.получить_текстуру(имя_текстуры);
            }
            catch (Throwable)
                в_журнал(имя_модуля, УровниЖурнала.Ошибка, 
                    "Ошибка получения текстуры для ЭлементИнфопанели с именем " ~ имя); 
            if (текстура !is null)
                результат.задать_текстуру(текстура);
        }
        if ("координаты_текстуры" in json_значение1) {
            try {
                координаты_текстуры = json_значение1["координаты_текстуры"].array;
                результат.задать_координаты_текстуры(координаты_текстуры[0].floating, координаты_текстуры[1].floating, 
                                    координаты_текстуры[2].floating, координаты_текстуры[3].floating);
            }
            catch (Throwable)
                {в_журнал(имя_модуля, УровниЖурнала.Ошибка, 
                        "Ошибка получения координат текстуры для ЭлементИнфопанели с именем " ~ имя); 
                }
        }
        if (будет_контейнером && ("элементы" in json_значение1))
            if (json_значение1["элементы"].type == JSON_TYPE.ARRAY) {
                foreach (JSONValue json_значение2; json_значение1["элементы"].array) {
                    ЭлементИнфопанели элемент = ЭлементИнфопанели.из_JSON(json_значение2, менеджер0);
                    if (элемент) {
                        результат.присоединить_элемент(элемент);
                    }
                }
            }
        return результат;
    }
}
