/++
Модуль, представляющий две структуры матриц: 
$(UL  $(LI  $(LINK2 #Матрица3, Матрица3) размерности 3x3  ) $(LI  $(LINK2 #Матрица4, Матрица4) размерности 4x4 )  )
+/
module ddd.matrix;

import std.string;
import ddd.vector, ddd.quaternion;

///Структура, представляющая матрицу размерности 3x3. Обычно используется в качестве матрицы поворота.
public struct Матрица3 {
    //------------Внимание!-------------
    // Матрица транспонирована!
    // Строки - это столбцы и наоборот!
    // А может и нет... Вобщем внимание!
    
    ///9 элементов матрицы, представленные в виде одномерного массива.
    public float[9] m; 

    ///Функция, используемая для вывода элементов матрицы в строковое представление
    public string toString(){
        return format("Матрица3(%(%6.4g, %))",m);
    }
}

/++
Структура, представляющая матрицу размерности 4x4. Используется для афинных преобразований,
т.е. задаёт поворот, сдвиг и масштаб в OpenGL.
+/
public struct Матрица4 {
    ///16 элементов матрицы, представленные в виде одномерного массива.
    public float[16] m; 
    
    //public const НУЛЕВАЯ_МАТРИЦА = Матрица4(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    //public this(){}

    /++
    Конструктор, получающий 16 чисел в качестве элементов матрицы
    
    Params:
        mij = элемент, находящийся в i-ой строке, j-ом столбце
    +/
    public this(
            float m00, float m01, float m02, float m03,
            float m10, float m11, float m12, float m13,
            float m20, float m21, float m22, float m23,
            float m30, float m31, float m32, float m33 )
        {
            m[0+0*4] = m00;
            m[0+1*4] = m01;
            m[0+2*4] = m02;
            m[0+3*4] = m03;
            m[1+0*4] = m10;
            m[1+1*4] = m11;
            m[1+2*4] = m12;
            m[1+3*4] = m13;
            m[2+0*4] = m20;
            m[2+1*4] = m21;
            m[2+2*4] = m22;
            m[2+3*4] = m23;
            m[3+0*4] = m30;
            m[3+1*4] = m31;
            m[3+2*4] = m32;
            m[3+3*4] = m33;
        }

    ///Конструктор копирования
    public this(const Матрица4 другая_матрица) {
        m=другая_матрица.m.dup;
    }

    /++
    Конструктор, строящий матрицу 4x4 на основе матрицы поворота 3x3 с нулевым сдвигом.
    
    Params:
        m3x3 = матрица поворота
    +/
    public this(const Матрица3 m3x3) {
        // Set up matrix with scale and rotation
        m[0+0*4] = m3x3.m[0+0*3]; m[0+1*4] = m3x3.m[0+1*3]; m[0+2*4] = m3x3.m[0+2*3]; 
        m[1+0*4] = m3x3.m[1+0*3]; m[1+1*4] = m3x3.m[1+1*3]; m[1+2*4] = m3x3.m[1+2*3]; 
        m[2+0*4] = m3x3.m[2+0*3]; m[2+1*4] = m3x3.m[2+1*3]; m[2+2*4] = m3x3.m[2+2*3]; 

        // Без сдвига
        m[0+3*4] = 0; m[1+3*4] = 0; m[2+3*4] = 0;
        // No projection term
        m[3+0*4] = 0; m[3+1*4] = 0; m[3+2*4] = 0; m[3+3*4] = 1;
    }
    
    /++
    Функция, заполняющая элементы матрицы для выполнения афинных преобразований.
    
    Порядок применения преобразований:
      1. Масштаб
      2. Поворот
      3. Сдвиг
    
    Params:
        сдвиг = вектор, на который выполняется сдвиг
        масштаб = вектор, элементы которого представляют коэффициенты масштаба по соответствующим осям
        поворот = кватернион, представляющий угол и ось поворота
    +/
    void матрица_преобразований(const(Вектор3) сдвиг, const(Вектор3) масштаб, const(Кватернион) поворот)
    {


        //Матрица3 матрица_поворота3x3=ориентация.в_матрицу_поворота();
        float[] m3x3=поворот.в_матрицу_поворота().m;

        //orientation.ToRotationMatrix(m3x3);

        // Получаем окончательную матрицу с масштабом, вращением и сдвигом
        m[0+0*4] = масштаб.x * m3x3[0+0*3]; m[0+1*4] = масштаб.y * m3x3[0+1*3]; 
        m[0+2*4] = масштаб.z * m3x3[0+2*3]; m[0+3*4] = сдвиг.x;
        
        m[1+0*4] = масштаб.x * m3x3[1+0*3]; m[1+1*4] = масштаб.y * m3x3[1+1*3]; 
        m[1+2*4] = масштаб.z * m3x3[1+2*3]; m[1+3*4] = сдвиг.y;
        
        m[2+0*4] = масштаб.x * m3x3[2+0*3]; m[2+1*4] = масштаб.y * m3x3[2+1*3]; 
        m[2+2*4] = масштаб.z * m3x3[2+2*3]; m[2+3*4] = сдвиг.z;

        // No projection term
        m[3+0*4] = 0; m[3+1*4] = 0; m[3+2*4] = 0; m[3+3*4] = 1;
    }

    /// Возвращает матрицу, полученную в результате транспонирования этой матрицы
    public Матрица4 транспонировать() 
    {
        return Матрица4(m[0+0*4], m[1+0*4], m[2+0*4], m[3+0*4],
                       m[0+1*4], m[1+1*4], m[2+1*4], m[3+1*4],
                       m[0+2*4], m[1+2*4], m[2+2*4], m[3+2*4],
                       m[0+3*4], m[1+3*4], m[2+3*4], m[3+3*4]);
    }

    /++
    Сложение, вычитание и умножение этой матрицы и матрицы m2

    Notes:
        Операция умножения не коммутативна, т.е. M1*M2 ≠ M2*M1
    +/
    public Матрица4 opBinary(string операция)(const(Матрица4) m2) 
        if (операция == "+" || операция == "-" || операция == "*")
    {
        Матрица4 r = Матрица4();
        static if (операция == "+") {
            r.m[0+0*4] = m[0+0*4] + m2.m[0+0*4]; r.m[0+1*4] = m[0+1*4] + m2.m[0+1*4]; 
            r.m[0+2*4] = m[0+2*4] + m2.m[0+2*4]; r.m[0+3*4] = m[0+3*4] + m2.m[0+3*4]; 
            r.m[1+0*4] = m[1+0*4] + m2.m[1+0*4]; r.m[1+1*4] = m[1+1*4] + m2.m[1+1*4]; 
            r.m[1+2*4] = m[1+2*4] + m2.m[1+2*4]; r.m[1+3*4] = m[1+3*4] + m2.m[1+3*4]; 
            r.m[2+0*4] = m[2+0*4] + m2.m[2+0*4]; r.m[2+1*4] = m[2+1*4] + m2.m[2+1*4]; 
            r.m[2+2*4] = m[2+2*4] + m2.m[2+2*4]; r.m[2+3*4] = m[2+3*4] + m2.m[2+3*4]; 
            r.m[3+0*4] = m[3+0*4] + m2.m[3+0*4]; r.m[3+1*4] = m[3+1*4] + m2.m[3+1*4]; 
            r.m[3+2*4] = m[3+2*4] + m2.m[3+2*4]; r.m[3+3*4] = m[3+3*4] + m2.m[3+3*4]; 
        }
        static if (операция == "-") {
            r.m[0+0*4] = m[0+0*4] - m2.m[0+0*4]; r.m[0+1*4] = m[0+1*4] - m2.m[0+1*4]; 
            r.m[0+2*4] = m[0+2*4] - m2.m[0+2*4]; r.m[0+3*4] = m[0+3*4] - m2.m[0+3*4]; 
            r.m[1+0*4] = m[1+0*4] - m2.m[1+0*4]; r.m[1+1*4] = m[1+1*4] - m2.m[1+1*4]; 
            r.m[1+2*4] = m[1+2*4] - m2.m[1+2*4]; r.m[1+3*4] = m[1+3*4] - m2.m[1+3*4]; 
            r.m[2+0*4] = m[2+0*4] - m2.m[2+0*4]; r.m[2+1*4] = m[2+1*4] - m2.m[2+1*4]; 
            r.m[2+2*4] = m[2+2*4] - m2.m[2+2*4]; r.m[2+3*4] = m[2+3*4] - m2.m[2+3*4]; 
            r.m[3+0*4] = m[3+0*4] - m2.m[3+0*4]; r.m[3+1*4] = m[3+1*4] - m2.m[3+1*4]; 
            r.m[3+2*4] = m[3+2*4] - m2.m[3+2*4]; r.m[3+3*4] = m[3+3*4] - m2.m[3+3*4]; 
        }
        static if (операция == "*") {
        r.m[0+0*4] = m[0+0*4] * m2.m[0+0*4] + m[0+1*4] * m2.m[1+0*4] + m[0+2*4] * m2.m[2+0*4] + m[0+3*4] * m2.m[3+0*4];
        r.m[0+1*4] = m[0+0*4] * m2.m[0+1*4] + m[0+1*4] * m2.m[1+1*4] + m[0+2*4] * m2.m[2+1*4] + m[0+3*4] * m2.m[3+1*4];
        r.m[0+2*4] = m[0+0*4] * m2.m[0+2*4] + m[0+1*4] * m2.m[1+2*4] + m[0+2*4] * m2.m[2+2*4] + m[0+3*4] * m2.m[3+2*4];
        r.m[0+3*4] = m[0+0*4] * m2.m[0+3*4] + m[0+1*4] * m2.m[1+3*4] + m[0+2*4] * m2.m[2+3*4] + m[0+3*4] * m2.m[3+3*4];

        r.m[1+0*4] = m[1+0*4] * m2.m[0+0*4] + m[1+1*4] * m2.m[1+0*4] + m[1+2*4] * m2.m[2+0*4] + m[1+3*4] * m2.m[3+0*4];
        r.m[1+1*4] = m[1+0*4] * m2.m[0+1*4] + m[1+1*4] * m2.m[1+1*4] + m[1+2*4] * m2.m[2+1*4] + m[1+3*4] * m2.m[3+1*4];
        r.m[1+2*4] = m[1+0*4] * m2.m[0+2*4] + m[1+1*4] * m2.m[1+2*4] + m[1+2*4] * m2.m[2+2*4] + m[1+3*4] * m2.m[3+2*4];
        r.m[1+3*4] = m[1+0*4] * m2.m[0+3*4] + m[1+1*4] * m2.m[1+3*4] + m[1+2*4] * m2.m[2+3*4] + m[1+3*4] * m2.m[3+3*4];

        r.m[2+0*4] = m[2+0*4] * m2.m[0+0*4] + m[2+1*4] * m2.m[1+0*4] + m[2+2*4] * m2.m[2+0*4] + m[2+3*4] * m2.m[3+0*4];
        r.m[2+1*4] = m[2+0*4] * m2.m[0+1*4] + m[2+1*4] * m2.m[1+1*4] + m[2+2*4] * m2.m[2+1*4] + m[2+3*4] * m2.m[3+1*4];
        r.m[2+2*4] = m[2+0*4] * m2.m[0+2*4] + m[2+1*4] * m2.m[1+2*4] + m[2+2*4] * m2.m[2+2*4] + m[2+3*4] * m2.m[3+2*4];
        r.m[2+3*4] = m[2+0*4] * m2.m[0+3*4] + m[2+1*4] * m2.m[1+3*4] + m[2+2*4] * m2.m[2+3*4] + m[2+3*4] * m2.m[3+3*4];

        r.m[3+0*4] = m[3+0*4] * m2.m[0+0*4] + m[3+1*4] * m2.m[1+0*4] + m[3+2*4] * m2.m[2+0*4] + m[3+3*4] * m2.m[3+0*4];
        r.m[3+1*4] = m[3+0*4] * m2.m[0+1*4] + m[3+1*4] * m2.m[1+1*4] + m[3+2*4] * m2.m[2+1*4] + m[3+3*4] * m2.m[3+1*4];
        r.m[3+2*4] = m[3+0*4] * m2.m[0+2*4] + m[3+1*4] * m2.m[1+2*4] + m[3+2*4] * m2.m[2+2*4] + m[3+3*4] * m2.m[3+2*4];
        r.m[3+3*4] = m[3+0*4] * m2.m[0+3*4] + m[3+1*4] * m2.m[1+3*4] + m[3+2*4] * m2.m[2+3*4] + m[3+3*4] * m2.m[3+3*4];
        }
        return r;
    }

    /// Умножение элементов этой матрицы на число
    public Матрица4 opBinary(string операция)(const float скаляр) 
        if (операция == "*")
        {
            return Матрица4(
                скаляр*m[0+0*4], скаляр*m[0+1*4], скаляр*m[0+2*4], скаляр*m[0+3*4],
                скаляр*m[1+0*4], скаляр*m[1+1*4], скаляр*m[1+2*4], скаляр*m[1+3*4],
                скаляр*m[2+0*4], скаляр*m[2+1*4], скаляр*m[2+2*4], скаляр*m[2+3*4],
                скаляр*m[3+0*4], скаляр*m[3+1*4], скаляр*m[3+2*4], скаляр*m[3+3*4]);
        }

    /// Возвращает вектор, на который производит сдвиг эта матрица
    public Вектор3 Позиция(){
        return Вектор3(m[0+3*4],m[1+3*4],m[2+3*4]);
    }

    /// Вывод элементов матрицы в строковое представление
    public string toString(){
        return format("Матрица4(%(%.4g, %))",m);
    }

    /// Возвращает минор этой матрицы размера 3x3, представляющий матрицу вращения и масштабирования
    Матрица3 получить_матрицу_3x3() {
        Матрица3 m3x3 = Матрица3();
        m3x3.m[0+0*3] = m[0+0*4];
        m3x3.m[0+1*3] = m[0+1*4];
        m3x3.m[0+2*3] = m[0+2*4];
        m3x3.m[1+0*3] = m[1+0*4];
        m3x3.m[1+1*3] = m[1+1*4];
        m3x3.m[1+2*3] = m[1+2*4];
        m3x3.m[2+0*3] = m[2+0*4];
        m3x3.m[2+1*3] = m[2+1*4];
        m3x3.m[2+2*3] = m[2+2*4];
        return m3x3;
    }

    const float
        MINOR(const Матрица4 mat, const size_t r0, const size_t r1, const size_t r2, 
                                const size_t c0, const size_t c1, const size_t c2)
    {
        float[] m=mat.m.dup;
        return m[r0+c0*4] * (m[r1+c1*4] * m[r2+c2*4] - m[r2+c1*4] * m[r1+c2*4]) -
            m[r0+c1*4] * (m[r1+c0*4] * m[r2+c2*4] - m[r2+c0*4] * m[r1+c2*4]) +
            m[r0+c2*4] * (m[r1+c0*4] * m[r2+c1*4] - m[r2+c0*4] * m[r1+c1*4]);
    }


    Матрица4 adjoint() const
    {
        return Матрица4( MINOR(this, 1, 2, 3, 1, 2, 3),
            -MINOR(this, 0, 2, 3, 1, 2, 3),
            MINOR(this, 0, 1, 3, 1, 2, 3),
            -MINOR(this, 0, 1, 2, 1, 2, 3),

            -MINOR(this, 1, 2, 3, 0, 2, 3),
            MINOR(this, 0, 2, 3, 0, 2, 3),
            -MINOR(this, 0, 1, 3, 0, 2, 3),
            MINOR(this, 0, 1, 2, 0, 2, 3),

            MINOR(this, 1, 2, 3, 0, 1, 3),
            -MINOR(this, 0, 2, 3, 0, 1, 3),
            MINOR(this, 0, 1, 3, 0, 1, 3),
            -MINOR(this, 0, 1, 2, 0, 1, 3),

            -MINOR(this, 1, 2, 3, 0, 1, 2),
            MINOR(this, 0, 2, 3, 0, 1, 2),
            -MINOR(this, 0, 1, 3, 0, 1, 2),
            MINOR(this, 0, 1, 2, 0, 1, 2));
    }

    /// Возвращает определитель матрицы
    float определитель() const
    {
        return m[0+0*4] * MINOR(this, 1, 2, 3, 1, 2, 3) -
            m[0+1*4] * MINOR(this, 1, 2, 3, 0, 2, 3) +
            m[0+2*4] * MINOR(this, 1, 2, 3, 0, 1, 3) -
            m[0+3*4] * MINOR(this, 1, 2, 3, 0, 1, 2);
    }

    /++ 
    Возвращает матрицу, обратную этой

    Note:
        Если определитель матрицы равен 0, элементы матрицы-результата будут равны $(B inf) (бесконечность)
    +/
    Матрица4 обратная() const
    {
        return adjoint() * (1.0f / определитель());
    }

}
