/++
Модуль, определяющий несколько функций для работы окном программы (приложения), а также с экраном в целом.
+/
module ddd.okno;

import std.conv;
import ddd.zavisimost.okno_SDL2, ddd.types;
import ddd.nastroyki_ddd;

/++
Функция, инициализирующая окно программы.

Параметры окна берутся из настроечного файла "настройки_ddd.cfg", или из настроек по-умолчанию,
см. описание модуля $(LINK2 nastroyki_ddd.html, nastroyki_ddd).
+/
bool инициализация_окна() {
    ПараметрыОкна параметры_окна;
    параметры_окна.ширина_окна = to!int(настройки.размеры_окна[0]);
    параметры_окна.высота_окна = to!int(настройки.размеры_окна[1]);
    параметры_окна.флаг_полный_экран = to!uint(настройки.полный_экран);
    параметры_окна.флаг_синхронизация = to!int(настройки.синхронизация_экрана);
    параметры_окна.режим_desktop = to!int(настройки.режим_desktop);
    параметры_окна.имя_окна = настройки.имя_окна;
    
    return инициализация_окна_SDL_GL(параметры_окна);  
}    

/++
Функция, закрывающая окно программы, т.е. завершающая её.
+/
void закрыть_окно() {
    SDL_выход();
}

/++
Функция, восстанавливающая состояние окна программы после минимизации.
+/
void восстановление_размеров_окна() {
    SDL_восстановление_размеров_окна();
}

/++
Функция, выводящая на экран содержимое ранее отрисованного буфера, вместо предыдущего буфера.

Обычно пользователю не требуется использовать эту функцию, её вызывает 
$(LINK2 vis.html#Визуализатор, Визуализатор) при реализации Игрового Цикла.
+/
void обменять_окна() {
    SDL_обменять_окна();
}

/++
Функция, возвращающая ширину и высоту окна в пикселях в формате структуры
$(LINK2 types.html#Координаты2D, Координаты2D).
+/
auto размеры_окна() {
    Координаты2D!int результат = Координаты2D!int(SDL_ширина_окна(), SDL_высота_окна());
    return результат;
}

/++
Возвращает состояние минимизации окна.

Returns:
    true, если окно было минимизировано, false в противном случае.
+/
bool окно_минимизировано() {
    return SDL_окно_минимизировано();
}    

/++
Устанавливает курсор мыши в центре окна.
+/
void установить_мышь_в_центре() {
    SDL_установить_мышь_в_центре();
}

/++
Выводит в журнал все доступные режимы экрана в системе.

Журнал должен быть заранее проинициализирован.
+/
void режимы_экрана_в_журнал() {
    SDL_режимы_экрана_в_журнал();
}

/++
Изменение режима экрана.

Params:
    номер_режима = Номер устанавливаемого режима. 
+/
void изменить_режим_экрана(int номер_режима) {
    SDL_изменить_режим_экрана(SDL_получить_окно(), номер_режима);
}
