/++
Главный модуль библиотеки, определяющий функцию инициализации главного окна и менеджера ddd, 
а также функцию, запускающую Игровой цикл.
В модуле присутствует экземпляр $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)а, через который пользователь будет получать
требуемые ему объекты и определять поведение программы.
+/
module ddd;

import std.string;
import ddd.nastroyki_ddd, ddd.okno, ddd.ddd_manager;
import my_utils.journal;

private enum имя_модуля = "ddd";
    
bool окно_инициализировано = false;

/++
Экземпляр менеджера, через который рекомендуется получать
объекты и определять поведение программы.
+/
DDDМенеджер менеджер;

/++
Функция инициализации главного окна и менеджера ddd.

Params:
    каталоги_ресурсов = Массив строк с путями к каталогам с ресурсами. 
    имя_файла_настроек = Название файла с настройками ddd. Если параметр отсутствует, то 
        используется имя файла $(B настройки_ddd.cfg), определённое в константе 
        $(LINK2 nastroyki_ddd.html#ИМЯ_ФАЙЛА_С_НАСТРОЙКАМИ_DDD, ИМЯ_ФАЙЛА_С_НАСТРОЙКАМИ_DDD).
Returns:
    true, если главное окно успешно инициализировано, false в противном случае.
+/
bool инициализация_ddd(string[] каталоги_ресурсов, string имя_файла_настроек = "") {
    string действительное_имя_файла_настроек = (имя_файла_настроек.empty) ? ИМЯ_ФАЙЛА_С_НАСТРОЙКАМИ_DDD : имя_файла_настроек;
    настройки.прочитать_настройки_из_файла(действительное_имя_файла_настроек);
    if (настройки.ошибочные_переменные.length) 
        в_журнал(имя_модуля, УровниЖурнала.Предупреждение, 
            format("Незагруженные переменные конфигурации ddd: %s" , настройки.ошибочные_переменные));
    в_журнал(имя_модуля, УровниЖурнала.Отладка,  
                журнал_многострочное_форматирование("настройки ddd = \n" ~ настройки.toString()));    
    окно_инициализировано = инициализация_окна();
    if (окно_инициализировано) {
        менеджер = new DDDМенеджер(каталоги_ресурсов);
    }
    return окно_инициализировано;
}

/++
Функция, запускающая Игровой цикл.

Params:
    максимум_кадров = Если задано положительное число, то будет визуализировано требуемое количество кадров,
        после чего программа завершится (такое поведение может использоваться в целях тестирования). 
        По-умолчанию цикл визуализации бесконечен.
+/
bool запустить_визуализацию(long максимум_кадров=-1) {
    менеджер.визуализатор.визуализация(максимум_кадров);
    закрыть_окно();
    return true;
}
