/++
Модуль, определяющий класс $(LINK2 #ТекстовыйЭлементИнфопанели, ТекстовыйЭлементИнфопанели), который позволяет 
добавлять текст на существующую $(LINK2 infopanel.html#Инфопанель, Инфопанель).
+/
module ddd.text_infopaneli;

import ddd.element_infopaneli, ddd.texture, ddd.ddd_manager;
import ddd.zavisimost.texture_SDL2_GL, ddd.zavisimost.shrift_SDL2;

import std.stdio, std.string;

import my_utils.journal;

enum имя_модуля = "текст_инфоп.";

/++
Класс, позволяющий
добавлять текст на существующую $(LINK2 infopanel.html#Инфопанель, Инфопанель).

Т.к. этот класс является потомком класса $(LINK2 element_infopaneli.html#ЭлементИнфопанели, ЭлементИнфопанели),
то все публичные методы класса $(B ЭлементИнфопанели) тоже действуют. Также это означает,
что объекту этого класса можно присвоить обычную текстуру вместо текста,
 и тогда такой элемент не будет отличаться от объекта класса $(B ЭлементИнфопанели).
+/
class ТекстовыйЭлементИнфопанели : ЭлементИнфопанели {
    private Текстура_SDL2_GL текстура_SDL = null;

    /++
    Конструктор, учитывающий координаты в относительных единицах.


    Отрицательные координаты x0 и/или y0 задают положение от конца инфопанели или 
    родительского контейнера (x0 от правого края, y0 от верхнего).
    


    Params:
        имя0 = Имя создаваемого элемента.
        будет_контейнером = Возможно ли вставлять в этот элемент другие элементы.
        x0 = Координата x левой границы элемента.
        y0 = Координата y нижней границы элемента.
        ширина0 = Ширина элемента.
        высота0 = Высота элемента.
    +/
    this(string имя0, bool будет_контейнером, float x0, float y0, float ширина0, float высота0) {
        super(имя0, будет_контейнером, x0, y0, ширина0, высота0);
    }

    /++
    Конструктор, принимающий минимальное количество параметров.


    Элемент, созданный этим конструктором, будет занимать всю площадь инфопанели или родительского контейнера.


    Params:
        имя0 = Имя создаваемого элемента.
        будет_контейнером = Возможно ли вставлять в этот элемент другие элементы.
    +/
    this(string имя0, bool будет_контейнером) {
        super(имя0, будет_контейнером);
    }

    /++
    Функция, позволяющая вывести (или изменить) текст на этом элементе.

    Params:
        менеджер = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)
        текст = текст, который требуется вывести в элементе
        имя_шрифта = имя ресурса со шрифтом для выводимого текста
        цвет = цвет текста в формате массива из 3-х значений RGB
    +/
    void задать_текст(DDDМенеджер менеджер, string текст, string имя_шрифта, float[3] цвет) {
        TTFШрифт шрифт = менеджер.получить_ttf_шрифт(имя_шрифта);
        if (шрифт) {
            инициализация_координат_текстуры();
            if (текстура is null) {
                текстура_SDL = new Текстура_SDL2_GL();
                текстура_SDL.построить_из_текста(шрифт, текст, цвет);
                задать_текстуру(текстура_SDL);
            }
            else {
                текстура_SDL.построить_из_текста(шрифт, текст, цвет);
            }
            
        }
    }

    /++
    Функция, изменяющая размеры элемента инфопанели, подгоняя их под размеры текущей текстуры.
    Это может потребоваться, чтобы текст, визуализированный в текстуру, размеры которой
    не были известны заранее, не искажался.

    Если у элемента не задана текстура (или не задан текст), или у элемента отсутсвует 
    родительский элемент, то функция не выполняет никаких действий.
    +/
    void обновить_размеры() {
        if ((текстура is null) || (_родитель is null)) {
            в_журнал(имя_модуля, УровниЖурнала.Предупреждение, 
                     format("Изменение размеров не получилось %s текстура=%s, _родитель=%s", имя, текстура, _родитель)); 
            return;
        }
        int ширина1 = текстура.получить_ширину();
        int высота1 = текстура.получить_высоту();
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("Изменение размеров %s ширина=%s, высота=%s", имя, ширина1, высота1)); 
        задать_абсолютные_размеры(ширина1, высота1);
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка,  
            format("Изменили размеры %s отн.x1=%s, отн.y1=%s, отн.x2=%s, отн.y2=%s",
                    имя,  относительный_x1, относительный_y1, относительный_x2, относительный_y2)); 
    }

}
