/++
Модуль, определяющий абстрактный класс $(LINK2 #Текстура, Текстура), представляющий текстуру 3D-объекта в сцене.
Реальный (не абстрактный) класс-потомок этого класса, для которого можно создать экземпляр - 
это класс $(B Текстура_SDL2_GL) из модуля $(B ddd.zavisimost.texture_SDL2_GL).

Также определён класс ресурсов $(LINK2 #ЗагруженныеТекстуры, ЗагруженныеТекстуры), 
позволяющий получать объекты-потомки
класса $(LINK2 #Текстура, Текстура) из файлов-ресурсов с графическими файлами 
(в данный момент поддерживаются форматы ".jpg", ".png", ".tga", ".tif"). 
Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты-потомки класса $(LINK2 #Текстура, Текстура) через него.
+/
module ddd.texture;

//import derelict.opengl.gl, derelict3.sdl2.sdl, derelict3.sdl2.image, derelict3.sdl2.ttf;
import std.string, core.stdc.string, std.conv;
//import std.string : fromStringz;

import ddd.resources; //, font;
import ddd.zavisimost.texture_SDL2_GL;
import my_utils.journal;

private enum имя_модуля = "текстура";
enum ТИП_РЕСУРСА = "картинка";

/*
inout(char)[] fromStringz(inout(char)* cString) @system pure {
    return cString ? cString[0 .. strlen(cString)] : null;
}
*/

/++
Абстрактный класс, представляющий текстуру 3D-объекта в сцене.
+/
class Текстура {

protected:

    int ширина;
	int высота;
	int формат;
	//int mIformat;

	uint номер_текстуры=0;

public:   

    /// Возвращает ширину текстуры в пикселях.
	int получить_ширину() {
		return ширина;
	}
	
    /// Возвращает высоту текстуры в пикселях.
	int получить_высоту() {
		return высота;
	}

    /++
    Возвращает соотношение между шириной и высотой текстуры.

    Если высота равна $(B 0), возвращается значение $(B -1).
    +/
    float получить_соотношение() {
		if (высота > 0)
			return to!float(ширина) / высота;
		else
			return -1;
	}

    /++
    Загрузка текстуры из графического файла.

    Это абстрактный метод, в классе-потомке $(B Текстура_SDL2) он переопределён для 
    загрузки текстур с помощью SDL2.

    Params:
        имя_файла = Имя графического файла, из которого будет загружена текстура.
    +/
    abstract void загрузить_из_файла(string имя_файла); 

    /++
    Включение этой текстуры в системе OpenGL, что позволит отображать выводимые 
    3D-объекты с использованием этой текстуры.
    
    Это абстрактный метод, в классе-потомке $(B Текстура_SDL2_GL) он переопределён.
    +/
    abstract void рисовать(); 

}

/++
Класс-контейнер ресурсов объектов класса $(LINK2 #Текстура, Текстура) и его потомков.


Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты классов текстур, построенных из графических файлов, через него.
+/
class ЗагруженныеТекстуры : ЗагруженныеРесурсы!(Текстура) {
	
    /++
    Конструктор, принимающий ссылку на объект типа $(LINK2 resources.html#ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов.
    +/
    this(ИменаРесурсов имена_ресурсов0) {
        super(имена_ресурсов0);
    }
    
    /++
    Функция загрузки в контейнер одного файла с ресурсом. 




    Returns:
        true, если файл удачно загрузился, false в противном случае.
    +/
    override protected bool загрузить_один(string имя) {
        string имя_файла;
        Текстура новая_текстура;
        try 
            имя_файла = имена_ресурсов.получить_имя_файла_ресурса(имя, ТИП_РЕСУРСА);
        catch (РесурсыИсключение) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Отсутствует файл текстуры %s", имя));
            return false;
        }
        новая_текстура = new Текстура_SDL2_GL();
        try 
			новая_текстура.загрузить_из_файла(имя_файла);
			//новая_текстура.load_fromFile("..\\" ~ имя_файла);
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Ошибка в файле текстуры %s", имя_файла));
            return false;
        }
		ресурсы[имя] = новая_текстура;
        return true;
    }
    
}
