module ddd.vector;

import std.stdio, std.string, std.math, core.exception;

/++
Структура, представляющая 3-х мерный вектор (x, y, z)

+/
public struct Вектор3 {

    /// Нулевой вектор
    public static const Вектор3 НОЛЬ = Вектор3(0,0,0);
    
    /// Единичные вектора по каждой из координатных осей
    public static const Вектор3 ЕДИНИЦА_X = Вектор3(1,0,0);
    /// ditto
    public static const Вектор3 ЕДИНИЦА_Y = Вектор3(0,1,0);
    /// ditto
    public static const Вектор3 ЕДИНИЦА_Z = Вектор3(0,0,1);
    /// ditto
    public static const Вектор3 МИНУС_ЕДИНИЦА_X = Вектор3(-1,0,0);
    /// ditto
    public static const Вектор3 МИНУС_ЕДИНИЦА_Y = Вектор3(0,-1,0);
    /// ditto
    public static const Вектор3 МИНУС_ЕДИНИЦА_Z = Вектор3(0,0,-1);
    
    /// Вектор (1, 1, 1), используемый для равномерного масштабирования
    public static const Вектор3 ЕДИНИЦА_МАСШТАБА = Вектор3(1, 1, 1);

	public static const float EPSILON=1e-06f;
	
    // Координаты вектора
    public float x, y, z;
    /*
    public this(){
        x=y=z=0.0f;
     }
*/
    public static this (){
        
    }
     
    /// Конструктор, принимающий координаты
    public this( float fX, float fY, float fZ ){
    //writeln("Конструктор Вектор3", fX, fY, fZ);
    x=fX;
    y=fY;
    z=fZ;
    }
    
    /// Конструктор копирования
    public this( const(Вектор3) другой_вектор ){
    x=другой_вектор.x;
    y=другой_вектор.y;
    z=другой_вектор.z;
    }

    /// Конструктор, принимающий координаты в виде массива
    public this(float[] массив){
    	try {
    		x=массив[0];
    		y=массив[1];
    		z=массив[2];
    	}
    	catch (RangeError e) {
    		x=y=z=0.0f;
    	}
    }
    
    /++
    Конструктор, формирующий вектор с координатами, равными переданному числу
    (обычно используется для равномерного масштабирования)
    +/
    public this(float скаляр){
		x=y=z=скаляр;
    }
    
    /// Метод, возвращающий длину вектора
    const public float длина()  {
        return this.length();
    }

    /// ditto
    const public float length()  {
        return cast(float) sqrt( x * x + y * y + z * z );
    }

    /// Операция, применяемая к координатам векторов поэлементно 
    public Вектор3 opBinary(string операция)(const(Вектор3) другой_вектор) 
        if (операция == "+" || операция == "-" || операция == "*")
    {
        mixin("return Вектор3(this.x" ~ операция ~ "другой_вектор.x, " ~
                             "this.y" ~ операция ~ "другой_вектор.y, " ~ 
                             "this.z" ~ операция ~ "другой_вектор.z);" );
    }

    /// Поэлементные операции над векторами "на месте"
    public Вектор3 opOpAssign(string операция)(const(Вектор3) другой_вектор) 
        if (операция == "+" || операция == "-" || операция == "*")
    {
        mixin("this.x" ~ операция ~ "=другой_вектор.x; " ~
              "this.y" ~ операция ~ "=другой_вектор.y; " ~ 
              "this.z" ~ операция ~ "=другой_вектор.z;" );
        return this;
    }
    
    /// Операция умножения вектора на число
    public Вектор3 opBinary(string операция)(const(float) скаляр) 
        if (операция == "*")
    {
        return Вектор3(
            x * скаляр,
            y * скаляр,
            z * скаляр);
    }
    
    /// Операция, возвращающая вектор, противоположный этому
    public Вектор3 opUnary(string операция)() 
        if (операция == "-")
    {
        return Вектор3(-x, -y, -z);
    }
    
    /++ 
    Функция, вычисляющая расстояние между векторами, т.е.
    длину вектора, являющегося разницей между этим вектором и переданным
    +/
    public float расстояние(const(Вектор3) другой_вектор)  {
        return (this - другой_вектор).длина();
    }
    
    /// Скалярное произведение векторов
    public float скалярное_произведение(const(Вектор3) другой_вектор)  {
        return ( x * другой_вектор.x + y * другой_вектор.y + z * другой_вектор.z );
    }
    
    /++
    Функция нормализации, приводящая вектор к единичной длине.
    
    Note: 
        Если длина вектора равна нулю, он остаётся без изменений.
    
    Returns:
        Значение первоначальной длины вектора.
    +/
    public float нормализация(){
    	float старая_длина = this.длина();
    	if (старая_длина>0) {
    		float множитель = 1.0f / старая_длина;
            x *= множитель;
            y *= множитель;
            z *= множитель;
    	}
    	return старая_длина;
    }
    
    /// Функция, возвращающая векторное произведение этого вектора на переданный
    public Вектор3 векторное_произведение( const(Вектор3) другой_вектор ) 
    {
        return Вектор3(
            y * другой_вектор.z - z * другой_вектор.y,
            z * другой_вектор.x - x * другой_вектор.z,
            x * другой_вектор.y - y * другой_вектор.x);
    }
    
    /// Функция, возвращающая значение угла в радианах между этим вектором и переданным
    public float угол_между( const(Вектор3) другой_вектор)
	{
		float произведение_длин = this.length() * другой_вектор.length();

		// Divide by zero check
		if(произведение_длин < EPSILON)
			произведение_длин = EPSILON;

		float f = this.скалярное_произведение(другой_вектор) / произведение_длин;

		if (f>1.0) f = 1.0f;
		if (f<-1.0) f = -1.0f;
		
		return cast(float) acos(f);

	}
    
    /// Проверка, является ли этот вектор нулевым
    public bool нулевая_длина() 
    {
        float квадрат_длины = (x * x) + (y * y) + (z * z);
        return (квадрат_длины < (EPSILON * EPSILON));

    }
    
    /++
    Функция, возвращающая новый вектор единичной длины, направленный так же, как и этот вектор.
    
    Note: 
        Если длина вектора равна нулю, возвращается такой же нулевой вектор.
    
    Returns:
        Вектор единичной длины, направленный так же, как и этот вектор.
        
    See Also:
        $(LINK2 #нормализация, нормализация)
    +/
    public Вектор3 нормализованная_копия() 
    {
        Вектор3 результат = Вектор3(this);
        результат.нормализация();
        return результат;
    }

    const public string toString(){
        //return format("Вектор3(%.8g,%.8g,%.8g)",x,y,z);
        return format("Вектор3(%.4g,%.4g,%.4g)",x,y,z);
    }
    

}

void проверка_вектора() {
    Вектор3 v1 = Вектор3(0,1,2);
    Вектор3 v2 = Вектор3(0,1,1);
    writeln(v1);
    writeln(v1+v2);
    writeln(v1-v2);
    writeln(v1*v2);
    writeln(v1 * 3);
    writeln(v1.нормализованная_копия());
    //writeln(v1/v2);
}
